/*
**	wmcdrom 0.5
**
**	wmcdrom 0.1: (c) 2002 Brian Schau <brisse@ping.dk>
**	next versions: (c) 2004 Matteo Lucarelli <matteo.lucarelli@taolab.it>
**
**	wmcdrom is a dock application which provides a way to easily
**	mount and unmount removable filesystems.
**
**	See the file LICENSE included in this package for license terms.
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 2 of the License, or
**	(at your option) any later version.
**
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
**
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to:
**
**		Free Software Foundation, Inc.
**		59 Temple Place - Suite 330
**		Boston, MA 02111-1307
**		USA
*/
#include <sys/mount.h>
#include <fstab.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <mntent.h>
#include <libgen.h>
#include <X11/X.h>
#include <gdk-pixbuf/gdk-pixbuf-xlib.h>
#include "cdromm.xpm"
#include "cdromu.xpm"
#include "dvdm.xpm"
#include "dvdu.xpm"
#include "floppym.xpm"
#include "floppyu.xpm"
#include "usbm.xpm"
#include "usbu.xpm"
#include "wait.xpm"


/* macros */
#define PROGRAMNAME "wmcdrom"
#define USAGE "[-m mountcmd] [-u umountcmd] [-i cd|dvd|floppy|usb] [-e ejectcmd] [-f filemanagercmd] [-c] devicefile\n\n  Left clik first mounts then launch fileman (-c chdir to mount point)\n  Right click first unmount then eject\n"
#define MOUNTCMD "/bin/mount"
#define UMOUNTCMD "/bin/umount"
#define EJECTCMD "/usr/bin/eject"
#define FILEMANCMD "xterm -e mc /mnt/ &"
#define XMESSAGECMD "xmessage -center -default okay"
#define MAXLINELEN 128
#define DOCKWINDOWSIZE 64
#define ICONWINDOWSIZE 48
#define SEXIT(x) { fprintf(stderr, x); exit(1); }
typedef enum { DeviceNotMounted, DeviceMounted , unknown} deviceStatus;

/* protos */
static int resolveDevice(void);
static void alarmHandler(int);
static void gfxMagic(char **xpm, Pixmap *icon);
static void initDock(char* iconame);
static void eventLoop(void);
static deviceStatus mountStatus(void);
static void updateMountStatus(void);
static void flashButtonUp(void);
static void flashButtonDown(void);
static void leftButtonAction(void);
static void rightButtonAction(void);
static int commandDevice(char *);
void  parsecommandline(char *line, char **argv);
void childHandler(int);

/* globals */
static deviceStatus oldDeviceStatus=unknown;
static Display *display;
static Window dockWindow, dockIconWindow;
static Pixmap mountPix, unmountPix , waitPix ;
static char *device, *cmdEject, *cmdMount, *cmdUmount, *cmdFileman;
static char realDevice[PATH_MAX]; //resolve device's real name if it's a symlink
static int deviceLen;
static int cdMntPoint = 0;

/*
**	alarmHandler
**
**	Catches alarm signal & updates display.
*/
static void alarmHandler(int signal)
{
	resolveDevice();
	updateMountStatus();
	alarm(1);
}

/*
 * Collect child process status from 
 */
void childHandler(int signo)
{
	sigset_t mask;
	int status = 0;

	sigemptyset(&mask);
	sigaddset(&mask, SIGALRM);

	sigprocmask(SIG_BLOCK, &mask, NULL);
	wait(&status);
	sigprocmask(SIG_UNBLOCK, &mask, NULL);
	alarm(1);
};

/*
**	main
*/
int main(int argc,char *argv[])
{
	struct sigaction act;
	int opt;
	char* typeIco="cd";

	while ((opt=getopt(argc, argv, "ce:m:u:i:f:"))!=-1) {
		switch (opt) {
		case 'c':
			cdMntPoint = 1;
			break;
		case 'e':
			cmdEject=strdup(optarg);
			break;
		case 'm':
			cmdMount=strdup(optarg);
			break;
		case 'u':
			cmdUmount=strdup(optarg);
			break;
		case 'f':
			cmdFileman=strdup(optarg);
			break;
		case 'i':
			typeIco=strdup(optarg);
			break;
		default:
			fprintf(stderr, "\nUsage: %s %s", argv[0], USAGE);
			exit(1);
		}
	}

	if (cmdEject==NULL) cmdEject=EJECTCMD;

	if (cmdMount==NULL) cmdMount=MOUNTCMD;

	if (cmdUmount==NULL) cmdUmount=UMOUNTCMD;

	if (cmdFileman==NULL) cmdFileman=FILEMANCMD;

	if (argc-optind==1) {
		device=argv[optind];
		deviceLen=strlen(device);
	} else {
		fprintf(stderr, "Usage: %s %s", argv[0], USAGE);
		exit(1);
	}

	if (!resolveDevice()) {
		fprintf(stderr, "Usage: %s %s", argv[0], USAGE);
		exit(1);
	}

	fprintf(stderr, "%s -> %s\n", device, realDevice);

	initDock(typeIco);

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler=alarmHandler;
	sigemptyset(&act.sa_mask);
	sigaction(SIGALRM, &act, NULL);

	memset(&act, 0, sizeof(struct sigaction));
	act.sa_handler=childHandler;
	sigemptyset(&act.sa_mask);
	sigaction(SIGCHLD, &act, NULL);

	gdk_pixbuf_xlib_init(display, DefaultScreen(display));

	if (!strcmp(typeIco,"cd")){
		gfxMagic(cdromm_xpm, &mountPix);
		gfxMagic(cdromu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"dvd")){
		gfxMagic(dvdm_xpm, &mountPix);
		gfxMagic(dvdu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"floppy")){
		gfxMagic(floppym_xpm, &mountPix);
		gfxMagic(floppyu_xpm, &unmountPix);
	}else if (!strcmp(typeIco,"usb")){
		gfxMagic(usbm_xpm, &mountPix);
		gfxMagic(usbu_xpm, &unmountPix);
	}

	gfxMagic(wait_xpm, &waitPix);

	updateMountStatus();
	XMapRaised(display, dockWindow);
	XFlush(display);
	alarm(1);

	eventLoop();

	alarm(0);

	XCloseDisplay(display);
	exit(0);
}

// parse a sentence in single words
void  parsecommandline(char *line, char **argv)
{
     while (*line != '\0') {      
          while (*line == ' ' || *line == '\t' || *line == '\n') *line++ = '\0';     
          *argv++ = line;          
          while (*line != '\0' && *line != ' ' && *line != '\t' && *line != '\n') line++;             
     }
     *argv = '\0';
}

/*
**	gfxMagic
**
**	Create a new icon.
**
**	 -> xpm		Source xpm.
**	 -> icon	Where to store icon.
*/
static void
gfxMagic(char **xpm, Pixmap *icon)
{
	GdkPixbuf *image;
	Pixmap mask;

	if ((image=gdk_pixbuf_new_from_xpm_data((const char **)xpm))==NULL)
		SEXIT("Failed to create pixbuf\n");

	gdk_pixbuf_xlib_render_pixmap_and_mask(image, icon, &mask, 128);
	gdk_pixbuf_unref(image);
	XFreePixmap(display, mask);
}
	
/*
**	initDock
**
**	Create dock dockWindow.
*/
static void initDock(char* iconame)
{
	XClassHint *classHint;
	XWMHints *hints;
	int black;
	char tmp[256];

	if ((display=XOpenDisplay(""))==NULL) 
		SEXIT("Cannot open display\n")

	black=BlackPixel(display, DefaultScreen(display));

	dockWindow=XCreateSimpleWindow(display, DefaultRootWindow(display), 0, 0, DOCKWINDOWSIZE, DOCKWINDOWSIZE, 0, black, black);

	if ((classHint=XAllocClassHint())==NULL)
		SEXIT("Failed to allocate class hint\n");

	strncpy(tmp,PROGRAMNAME,256);
	strncat(tmp,iconame,256);

	classHint->res_class=tmp;
	classHint->res_name=tmp;
	XSetClassHint(display, dockWindow, classHint);
	XFree(classHint);

	dockIconWindow=XCreateSimpleWindow(display, dockWindow, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, (DOCKWINDOWSIZE-ICONWINDOWSIZE)/2, ICONWINDOWSIZE, ICONWINDOWSIZE, 0, black, black);

	if ((hints=XAllocWMHints())==NULL)
		SEXIT("Failed to allocate hints\n");

	hints->flags=StateHint|WindowGroupHint|IconWindowHint;
	hints->initial_state=WithdrawnState;
	hints->window_group=dockWindow;
	hints->icon_window=dockIconWindow;
	hints->icon_x=0;
	hints->icon_y=0;
	XSetWMHints(display, dockWindow, hints);
		
	XSelectInput(display, dockWindow, ExposureMask);
	XSelectInput(display, dockIconWindow, ButtonPressMask|ButtonReleaseMask|ExposureMask);
}

/*
**	eventLoop
**
**	Handle events from X server.
*/
static void eventLoop(void)
{
	XEvent ev;

	for (;;) {
		XNextEvent(display, &ev);

		switch (ev.type) {
		case Expose:
			updateMountStatus();
			break;
		case ButtonPress:
			if (ev.xbutton.x<0 || ev.xbutton.x>ICONWINDOWSIZE || ev.xbutton.y<0 || ev.xbutton.y>ICONWINDOWSIZE)
				break;
			flashButtonUp();
			break;
		case ButtonRelease:
			if (ev.xbutton.x<0 || ev.xbutton.x>ICONWINDOWSIZE || ev.xbutton.y<0 || ev.xbutton.y>ICONWINDOWSIZE)
				break;
			if (ev.xbutton.button==1){
				flashButtonDown();
				leftButtonAction();
			}else if (ev.xbutton.button==3){
				flashButtonDown();
				rightButtonAction();
			} else if (ev.xbutton.button == 2) {
				flashButtonDown();
				commandDevice(cmdEject);
			}
			break;
		case DestroyNotify:
			return;
		default:
			break;
		}
	}
}

/*
 * Recursively search for device real name
 */
static int resolveDevice(void)
{
	struct stat st;

	strncpy(realDevice, device, sizeof(realDevice));

	do {
		if (lstat(realDevice, &st) == -1)
			return 0;

		if (S_ISLNK(st.st_mode)) {
			char linkName[PATH_MAX];
			ssize_t linkLen;

			if ((linkLen = readlink(realDevice, linkName, sizeof(linkName))) > 0) {
				linkName[linkLen] = 0;

				if (linkName[0] != '/') {
					char *path = NULL;
					char *dir = NULL;

					if ((path = strdup(realDevice)) == NULL) {
						perror("strdup");
						exit(EXIT_FAILURE);
					}
					if ((dir = dirname(path)) == NULL) {
						perror("dirname");
						free(path);
						return 0;
					}

					strncpy(realDevice, dir, strlen(dir) + 1);
					strcat(realDevice, "/");
					strcat(realDevice, linkName);
					free(path);

				} else
					strncpy(realDevice, linkName, sizeof(realDevice));

			} else {
				perror("readlink");
				return 0;
			}

		} else if (S_ISBLK(st.st_mode)) {
			break;
		}

	} while (1);

	return 1;
}

/*
**	mountStatus
**
**	Return mount status for this device.
*/
static deviceStatus mountStatus(void)
{
	/* use a more portable way */
	FILE *fstab = NULL;
	struct mntent *ent = NULL;
	deviceStatus d = DeviceNotMounted;

	if ((fstab = setmntent("/etc/mtab", "r")) != NULL) {
		while ((ent = getmntent(fstab)) != NULL) {
			if ((strcmp(ent->mnt_fsname, realDevice) == 0) || (strcmp(ent->mnt_fsname, realDevice) == 0)) {
				d = DeviceMounted;
				break;
			}
		}
		endmntent(fstab);
	} else
		perror("setmntent");

	return d;
}

/*
**	updateMountStatus
**
**	Update graphics in dockWindow.
*/
static void updateMountStatus(void)
{
	deviceStatus d=mountStatus();

	if (d==oldDeviceStatus) return;

	oldDeviceStatus=d;

	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, (d==DeviceMounted) ? mountPix : unmountPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

/*
**	flashButton
**	Flash icon for click feedback
*/
static void flashButtonUp(void)
{
	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, waitPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}
static void flashButtonDown(void)
{
	deviceStatus d=mountStatus();

	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, (d==DeviceMounted) ? mountPix : unmountPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}
}

/*
**	leftButtonAction
**	Dispatch for left mousebutton click on icon.
*/
static void leftButtonAction(void)
{
	// mount or lounch fileman
	if (oldDeviceStatus==DeviceNotMounted)
		commandDevice(cmdMount);
	else if (fork() == 0) {
		char* argv[128];
		FILE *fstab = NULL;
		struct mntent *ent = NULL;

		parsecommandline(cmdFileman,argv);

		if (cdMntPoint) {
			if ((fstab = setmntent("/etc/fstab", "r")) != NULL) {
				while ((ent = getmntent(fstab)) != NULL) {
					if (strcmp(ent->mnt_fsname, device) == 0) {
						if (chdir(ent->mnt_dir) == -1)
							perror("chdir");
						break;
					}
				}
				endmntent(fstab);
			} else
				perror("setmntent");
		}

		if (execvp(*argv, argv) < 0) {
			printf("WMCDROM: launching fileman failed!\n");
			exit(1);
		}
	}

	mountStatus();
}

/*
**	rightButtonAction
**
**	Dispatch for right mousebutton click on icon.
*/
static void rightButtonAction(void)
{
	// umount or eject
	if (oldDeviceStatus==DeviceNotMounted) commandDevice(cmdEject);
	else commandDevice(cmdUmount);

	mountStatus();
}

/*
**  execute command on Device
**	 -> com		Command.
*/
static int commandDevice(char *com)
{
	int ret=-1;
	char* str=(char* )calloc(strlen(com)+deviceLen+7,1);
	FILE *filep = NULL;
	char com_out[getpagesize()];
	const int len = getpagesize();

	if (display) {
		XSetWindowBackgroundPixmap(display, dockIconWindow, waitPix);
		XClearWindow(display, dockIconWindow);
		XFlush(display);
	}

	if (str){
		sprintf(str,"%s %s 2>&1",com,device);
		filep = popen(str, "r");

		if (filep != NULL) {
			setbuf(filep, NULL);
			sleep(1); //wait stdout

			if (fgets(com_out, len, filep) != NULL) {
				char xmsg[strlen(com_out) + strlen(XMESSAGECMD) + 4];

				fprintf(stderr, "%s error: %s", com, com_out);
				sprintf(xmsg, "%s \"%s\"", XMESSAGECMD, com_out);
				system(xmsg);
			}

			pclose(filep);
		}

		free(str);
	}

	flashButtonDown();

	return ret;
}

